# $Id: snmp.tcl,v 2.42 2002/05/05 11:38:21 gdmr Exp $
#
# Copyright 2000 Division of Informatics, University of Edinburgh
# 
# Permission to use, copy, modify, distribute, and sell this software and its
# documentation for any purpose is hereby granted without fee, provided that
# the above copyright notice appear in all copies and that both that
# copyright notice and this permission notice appear in supporting
# documentation, and that the name of the University of Edinburgh not be used
# in advertising or publicity pertaining to distribution of the software
# without specific, written prior permission.  The University of Edinburgh
# makes no representations about the suitability of this software for any
# purpose.  It is provided "as is" without express or implied warranty.
#
# SNMP-based procedures to query the switches.  Cache the results
# where appropriate.
#

####################
# Initialisation
####################

puts "Loading Tnm..."
package require Tnm 3.0
namespace import Tnm::snmp
namespace import Tnm::mib

# To speed up loading we pare things down to the bare minimum
set tnm(mibs) \
   "MAU-MIB BRIDGE-MIB RMON-MIB EtherLike-MIB P-BRIDGE-MIB Q-BRIDGE-MIB"
set mibList \
   {hpconfig.mib hpvlan.mib hpicfoid.mib hpbasic.mib hpff.mib hpchassis.mib \
    stat.mib hphttpmg.mib}

#puts "MIBs: $mibList"
foreach m $mibList {
   mib load [file join [file dirname $argv0] $m]
}

set snmpRetries 1
set snmpTimeout 30

proc snmpInit {host community} {
   global snmpS
   global snmpRetries
   global snmpTimeout
   if [info exists snmpS] {
      puts "Can't re-snmpInit"
   } else {
      set snmpS [snmp generator \
                  -address $host -community $community \
                  -retries $snmpRetries -timeout $snmpTimeout]
   }
}

####################
# System group
####################

proc snmpDescr {} {
   global snmpS
   global snmpSysDescr
   if [info exists snmpSysDescr] {
      set snmpSysDescr
   } else {
      set vb "SNMPv2-MIB!sysDescr.0"
      set vb [$snmpS get $vb]
      eval set snmpSysDescr [snmp value $vb]
   }
}

proc snmpUpTime {} {
   # Don't cache this one
   global snmpS
   set vb "SNMPv2-MIB!sysUpTime.0"
   set vb [$snmpS get $vb]
   set v [snmp value $vb]
   # Divide by 100 by chopping the last two digits off the end of
   # the timeticks value...
   string range $v 0 [expr [string length $v] - 3]
}

proc snmpLocation {} {
   global snmpS
   global snmpSysLocation
   if [info exists snmpSysLocation] {
      set snmpSysLocation
   } else {
      set vb "SNMPv2-MIB!sysLocation.0"
      set vb [$snmpS get $vb]
      eval set snmpSysLocation [snmp value $vb]
   }
}

####################
# Interfaces, etc.
####################

proc snmpEtherIfIndices {} {
   global snmpS
   global ifNameToIndex
   global ifIndexToName
   set ifRoot [mib oid "IF-MIB!ifIndex"]
   set vbl "$ifRoot.0
            IF-MIB!ifType.0
            IF-MIB!ifDescr.0"
   while {1} {
      set vbl [$snmpS getnext $vbl]
      if {! [mib subtree $ifRoot [snmp oid $vbl 0]]} break
      if {[snmp value $vbl 1] == "ethernetCsmacd"} {
         lappend indices [snmp value $vbl 0]
         set i [snmp value $vbl 0]
         set n [snmp value $vbl 2]
         set ifNameToIndex($n) $i
         set ifIndexToName($i) $n
      }
   }
   # Don't need to sort, they'll be in numeric order already
   set indices
}

proc snmpIfName {i} {
   global ifIndexToName
   if [info exists ifIndexToName($i)] {
      set ifIndexToName($i)
   } else {
      return "*unknown $i*"
   }
}

proc snmpIfIndex {n} {
   global ifNameToIndex
   if [info exists ifNameToIndex($n)] {
      set ifNameToIndex($n)
   } else {
      return -1
   }
}
proc snmpIfType {i} {
   global snmpIfTypes
   global snmpS
   if {[info exists snmpIfTypes($i)]} {
      set snmpIfTypes($i)
   } else {
      set vb "IF-MIB!ifType.$i"
      set vb [$snmpS get $vb]
      set snmpIfTypes($i) [snmp value $vb]
   }
}

proc snmpIfDescr {i} {
   global snmpIfDescrs
   global snmpS
   if {[info exists snmpIfDescrs($i)]} {
      set snmpIfDescrs($i)
   } else {
      set vb "IF-MIB!ifDescr.$i"
      set vb [$snmpS get $vb]
      eval set snmpIfDescrs($i) [snmp value $vb]
   }
}

proc snmpIfUp {i} {
   global snmpS
   set vb "IF-MIB!ifOperStatus.$i"
   set vb [$snmpS get $vb]
   if {[snmp value $vb] == "up"} {
      return 1
   } else {
      return 0
   }
}

proc snmpIfAdmin {i} {
   global snmpS
   set vb "IF-MIB!ifAdminStatus.$i"
   set vb [$snmpS get $vb]
   if {[snmp value $vb] == "up"} {
      return 1
   } else {
      return 0
   }
}

proc snmpHpPortAdmin {i} {
   global snmpS
   set vb "CONFIG-MIB!hpSwitchPortAdminStatus.$i"
   set vb [$snmpS get $vb]
   if {[snmp value $vb] == "enable"} {
      return 1
   } else {
      return 0
   }
}

proc snmpIfSetAdmin {i {mode "up"}} {
   global snmpS
   puts "snmpIfSetAdmin: $i $mode"
   if {[string compare $mode "up"] && [string compare $mode "down"]} {
      puts "snmpIfSetAdmin: bad mode \"$mode\""
      return 0
   }
   set vbl [list [list "IF-MIB!ifAdminStatus.$i" "$mode"]]
   if [catch {$snmpS set $vbl} why] {
      puts "snmpIfSetAdmin: $why"
      return 0
   } else {
      puts "OK!"
      return 1
   }
}

#proc snmpHpSetPortAdmin {i {mode "enable"}} {
#   global snmpS
#   puts "snmpHpSetPortAdmin: $i $mode"
#   if {[string compare $mode "enable"] && [string compare $mode "disable"]} {
#      puts "snmpIfSetAdmin: bad mode \"$mode\""
#      return 0
#   }
#   set vbl [list [list "CONFIG-MIB!hpSwitchPortAdminStatus.$i" "$mode"]]
#   if [catch {$snmpS set $vbl} why] {
#      puts "snmpHpSetPortAdmin: $why"
#      return 0
#   } else {
#      puts "OK!"
#      return 1
#   }
#}


proc snmpIfSpeed {i} {
   global snmpS
   set vb "IF-MIB!ifSpeed.$i"
   set vb [$snmpS get $vb]
   snmp value $vb
}

proc snmpIfLastChange {i} {
   global snmpS
   set vb "IF-MIB!ifLastChange.$i"
   set vb [$snmpS get $vb]
   snmp value $vb
}

proc snmpIfStats {i} {
   global snmpS
   set vbl "IF-MIB!ifInOctets.$i
            IF-MIB!ifInUcastPkts.$i
            IF-MIB!ifInErrors.$i
            IF-MIB!ifOutOctets.$i
            IF-MIB!ifOutUcastPkts.$i
            IF-MIB!ifOutErrors.$i"
   set vbl [$snmpS get $vbl]
   snmp value $vbl
}

proc snmpIfTrunkGroup {i} {
   global snmpS
   set vb "CONFIG-MIB!hpSwitchPortTrunkGroup.$i"
   set vb [$snmpS get $vb]
   snmp value $vb
}

####################
# VLANs
####################

set qBridge 0
proc snmpVlanScan {} {
   global snmpS
   global qBridge
   global verbose
   global vlanTagToIndex
   global vlanIndexToTag
   global vlanIndexToName
   # There are two possibilities at this point.  Either we have a new(ish)
   # switch which knows about the Q-BRIDGE mib, in which case we use that,
   # or we have to use HP's old proprietary mechanism.  Try for the
   # standard first...
   set vlRoot [mib oid "Q-BRIDGE-MIB!dot1qVlanStaticName"]
   set vlMatch "^$vlRoot\\.(.*)$"
   set vbl "$vlRoot.0"
   set vbl [$snmpS getnext $vbl]
   if [mib subtree $vlRoot [snmp oid $vbl]] {
      # If the result we got back is part of the dot1qVlanStaticName
      # table then we're in Q-BRIDGE mode.
      if {$verbose} {puts "VLANs: Q-BRIDGE"}
      set qBridge 1
      if [regexp $vlMatch [snmp oid $vbl 0] all tag] {
         set name [snmp value $vbl]
         #puts "VLAN tag $tag, name $name"
         # The index is the tag, but we'll register it anyway
         set vlanTagToIndex($tag) $tag
         set vlanIndexToTag($tag) $tag
         set vlanIndexToName($tag) $name
      } else {
         puts stderr "snmpVlanScan: no regexp match for [snmp oid $vbl]"
      }
      # Now look to see if there are any more
      while {1} {
         set vbl [$snmpS getnext $vbl]
         if {![mib subtree $vlRoot [snmp oid $vbl]]} break
         if [regexp $vlMatch [snmp oid $vbl 0] all tag] {
            set name [snmp value $vbl]
            #puts "VLAN tag $tag, name $name"
            # The index is the tag, but we'll register it anyway
            set vlanTagToIndex($tag) $tag
            set vlanIndexToTag($tag) $tag
            set vlanIndexToName($tag) $name
         } else {
            puts stderr "snmpVlanScan: no regexp match for [snmp oid $vbl]"
         }
      }
   } else {
      # No dot1qVlanStaticName match, so we're in HP-VLAN mode
      if {$verbose} {puts "VLANs: HP-VLAN"}
      set vlRoot [mib oid "HP-VLAN!hpVlanIdentMode"]
      set vlMatch "^$vlRoot\\.(.*)$"
      set vbl "$vlRoot.0
               HP-VLAN!hpVlanIdentName.0
               HP-VLAN!hpVlanDot1QID.0"
      while {1} {
         set vbl [$snmpS getnext $vbl]
         if {![mib subtree $vlRoot [snmp oid $vbl 0]]} break
         if [regexp $vlMatch [snmp oid $vbl 0] all i] {
            set name [snmp value $vbl 1]
            set tag [snmp value $vbl 2]
            #puts "VLAN index $i, tag $tag, name $name"
            if [string compare [snmp value $vbl 0] "port"] {
               puts "VLAN $name ($tag) isn't port-based"
            } else {
               set vlanTagToIndex($tag) $i
               set vlanIndexToTag($i) $tag
               set vlanIndexToName($i) $name
            }
         } else {
            puts stderr "snmpVlanScan: no regexp match for [snmp oid $vbl 0]"
         }
      }
   }
   lsort -integer [array names vlanTagToIndex]
}

proc snmpVlanTagToName {tag} {
   global vlanTagToIndex
   global vlanIndexToName
   if [info exists vlanTagToIndex($tag)] {
      set vlanIndexToName($vlanTagToIndex($tag))
   } else {
      return "*unknown*"
   }
}

# Utility procedure to test a bit in a PortList string
proc qBtest {pl if} {
   #puts "Test bit $if of $pl"
   set bit [expr ($if - 1) & 7]
   set byte [expr ($if - 1) >> 3]
   set mask [expr 128 >> $bit]
   #puts "$if: check bit $bit (mask $mask) of byte $byte"
   set outcome [eval expr 0x[lindex [split $pl ":"] $byte] & $mask]
   #puts "Got $outcome"
   set outcome
}

# Utility procedure to clear a bit in a PortList string
proc qBclear {pl if} {
   #puts "Clear bit $if of $pl"
   set bit [expr ($if - 1) & 7]
   set byte [expr ($if - 1) >> 3]
   set mask [expr 128 >> $bit]
   set pll [split $pl ":"]
   #puts "$if: clear bit $bit (mask $mask) of byte $byte"
   set newValue [eval expr 0x[lindex $pll $byte] & ~$mask]
   set outcome [join [lreplace $pll $byte $byte [format "%02X" $newValue]] ":"]
   #puts "Outcome $outcome"
   set outcome
}

# Utility procedure to set a bit in a PortList string
proc qBset {pl if} {
   #puts "Set bit $if of $pl"
   set bit [expr ($if - 1) & 7]
   set byte [expr ($if - 1) >> 3]
   set mask [expr 128 >> $bit]
   set pll [split $pl ":"]
   #puts "$if: set bit $bit (mask $mask) of byte $byte"
   set newValue [eval expr 0x[lindex $pll $byte] | $mask]
   set outcome [join [lreplace $pll $byte $byte [format "%02X" $newValue]] ":"]
   #puts "Outcome $outcome"
   set outcome
}

proc snmpVlanLists {if} {
   global snmpS
   global qBridge
   global vlanIndexToTag
   global vlanStaticEgress
   global vlanStaticUntagged
   if {$qBridge} {
      # Q-BRIDGE
      # The first thing we need is a list of tags
      set tagList [lsort -integer [array names vlanIndexToTag]]
      # Now, have we grabbed the ports lists yet?  Assume that
      # we haven't if there isn't one defined for tag 1
      if {![info exists vlanStaticUntagged(1)]} {
         foreach t $tagList {
            #puts "Grabbing port lists for tag $t"
            set vbl "Q-BRIDGE-MIB!dot1qVlanStaticEgressPorts.$t
                     Q-BRIDGE-MIB!dot1qVlanStaticUntaggedPorts.$t"
            set vbl [$snmpS get $vbl]
            set vlanStaticEgress($t) [snmp value $vbl 0]
            set vlanStaticUntagged($t) [snmp value $vbl 1]
         }
         #foreach t [lsort -integer [array names vlanStaticEgress]] {
         #   puts "$t: E $vlanStaticEgress($t)"
         #   puts "$t: U $vlanStaticUntagged($t)"
         #}
      }
      set untagged 0
      set tagged {}
      foreach t [array names vlanStaticEgress] {
         if [qBtest $vlanStaticUntagged($t) $if] {
            set untagged $t
         } elseif [qBtest $vlanStaticEgress($t) $if] {
            lappend tagged $t
         }
      }
   } else {
      # HP-VLAN
      set indices [lsort -integer [array names vlanIndexToTag]]
      # XXX We could either batch up all the queries for the interface
      # and send them at once, or we could send them one at a time.
      # We'll do the latter here for now, though it may be more network-
      # intensive, as it means we don't have to extract the VLAN index
      # from the vbl entries later.
      set tagged {}
      foreach i $indices {
         set vb "HP-VLAN!hpVlanMemberTagged2.$i.$if"
         set vb [$snmpS get $vb]
         switch -exact -- [snmp value $vb] {
            "tagged"    {
               lappend tagged $vlanIndexToTag($i)
            }
            "untagged"  {
               if [info exists untagged] {
                  puts stderr "snmpVlanLists: $i.$if already untagged"
               }
               set untagged $vlanIndexToTag($i)
            }
            "no"        { }
            default     {
               puts stderr "snmpVlanLists: Unknown value for $i.$if"
            }
         }
      }
   }
   list $untagged [lsort -integer $tagged]
}

proc snmpCreateVlan {name tag} {
   global vlanTagToIndex
   global vlanIndexToTag
   global vlanIndexToName
   global vlanStaticEgress
   global vlanStaticUntagged
   global qBridge
   global snmpS
   puts "snmpCreateVlan: name $name, tag $tag"
   if {$qBridge} {
      # Q-BRIDGE
      # First check to make sure the vlan doesn't exist already.
      if [info exists vlanStaticEgress($tag)] {
         puts stderr "snmpCreateVlan: VLAN $tag already exists!"
         return 0
      }
      # Rows in the static table are indexed by tag, so just go ahead and
      # try to create a new one.
      set vbl [list \
               [list "Q-BRIDGE-MIB!dot1qVlanStaticName.$tag" "$name"] \
               [list "Q-BRIDGE-MIB!dot1qVlanStaticRowStatus.$tag" \
                     "createAndGo"]]
      if [catch {$snmpS set $vbl} why] {
         puts stderr "snmpCreateVlan: tag $tag: $why"
         return 0
      }
      # Looking OK, so try to grab the Egress/Untagged lists.
      # We could try to calculate this, but it's easier to ask
      # the switch how many zeros there should be.
      set vbl "Q-BRIDGE-MIB!dot1qVlanStaticEgressPorts.$tag
               Q-BRIDGE-MIB!dot1qVlanStaticUntaggedPorts.$tag"
      set vbl [$snmpS get $vbl]
      set vlanStaticEgress($tag) [snmp value $vbl 0]
      set vlanStaticUntagged($tag) [snmp value $vbl 1]
      #puts "E: $vlanStaticEgress($tag)"
      #puts "U: $vlanStaticUntagged($tag)"
      # Note name-tag translation
      set vlanTagToIndex($tag) $tag
      set vlanIndexToTag($tag) $tag
      set vlanIndexToName($tag) $name
      # The VLAN will come up with IP disabled, so that's all
      # we have to do here.
      puts "OK"
      return 1
   } else {
      # HP-VLAN
      # We have to iterate the first couple of stages of this, as if we've
      # previously deleted something from a slot it sometimes (always??!)
      # happens that the switch won't let us reuse it.
      set livesLeft 9
      set free 1
      set inUse [lsort -integer [array names vlanIndexToTag]]
      while {1} {
         # Stage 1: find a free slot in the table
         for {} {1} {incr free} {
            if {[lsearch -exact $inUse $free] < 0} { break }
         }
         # Stage 2: create the table entry
         set vbl [list \
                  [list "HP-VLAN!hpVlanIdentName.$free" "$name"] \
                  [list "HP-VLAN!hpVlanDot1QID.$free" "$tag"] \
                  [list "HP-VLAN!hpVlanIdentStatus.$free" "createAndGo"]]
         #puts stderr "$snmpS set $vbl"
         if [catch {set vbl [$snmpS set $vbl]} why] {
            puts "snmpCreateVlan: slot $free: $why"
            # OK, the create operation failed.  It could be because the
            # switch is in a huff about that particular slot, so if we have
            # any lives left we'll go and look for another one.
            if {[incr livesLeft -1] <= 0} {
               puts "snmpCreateVlan: no lives left"
               return 0
            }
            puts stderr "Retrying $livesLeft..."
            incr free
         } else {
            break
         }
      }
      set vlanTagToIndex($tag) $free
      set vlanIndexToTag($free) $tag
      set vlanIndexToName($free) $name
      # Stage 3: find the index for the switch's interface
      # to the VLAN
      set ifRoot [mib oid "IF-MIB!ifName"]
      set vbl [list [list $ifRoot.1]]
      if [catch {set vbl [$snmpS get $vbl]} why] {
         puts "Get: $why"
         puts "vbl: $vbl"
         set vbl [$snmpS get $vbl]
      }
      while {1} {
         if {![mib subtree $ifRoot [snmp oid $vbl]]} {
            puts stderr "snmpCreateVlan: Couldn't find ifName for $name"
            return 0
         }
         if {[string compare $name [snmp value $vbl]] == 0} {
            break
         }
         set vbl [$snmpS getnext $vbl]
      }
      if {![regexp "$ifRoot\.(.*)$" [snmp oid $vbl] all ifIndex]} {
         puts stderr "snmpCreateVlan: no regexp match for $vbl"
         return 0
      }
      # Stage 4: turn off dhcp on it
      set vbl [list \
               [list "HP-ICF-BASIC!hpicfIpConfigAdminStatus.$ifIndex" \
                     "disable"]]
      if [catch {$snmpS set $vbl} why] {
         puts stderr "snmpCreateVlan: $why"
         return 0
      }
      puts "OK!"
      return 1
   }
}

proc snmpDeleteVlan {name tag} {
   global qBridge
   global snmpS
   global vlanTagToIndex
   # Assumption: we'll never be called if there are still any
   # ports using the VLAN we're about to delete.
   puts stderr "snmpDeleteVlan: $name $tag"
   if {$qBridge} {
      # Q-BRIDGE
      set vbl [list [list "Q-BRIDGE-MIB!dot1qVlanStaticRowStatus.$tag" \
                          "destroy"]]
   } else {
      # HP-VLAN
      if {![info exists vlanTagToIndex($tag)]} {
         puts stderr "Can't find index for tag $tag??"
         return 0
      }
      set i $vlanTagToIndex($tag)
      set vbl [list [list "HP-VLAN!hpVlanIdentStatus.$i" "destroy"]]
   }
   if [catch {$snmpS set $vbl} why] {
      puts stderr "snmpDeleteVlan: $why"
      return 0
   }
   puts "OK!"
   return 1
}

proc snmpSetVlans {index label untagged tagged} {
   global qBridge
   global vlanTagToIndex
   global vlanIndexToTag
   global vlanIndexToName
   global vlanStaticEgress
   global vlanStaticUntagged
   global snmpS
   puts "snmpSetVlans: PI $index, PL $label, U $untagged, T $tagged"
   if {$qBridge} {
      # Q-BRIDGE
      # There are constraints on the combinations of ways that "egress" and
      # "untagged" are set: we can't have more than one VLAN set as
      # untagged for a port, and we can't have an untagged port that isn't
      # also in the egress set.  So, take three passes...
      # (It would perhaps be more efficient to batch up the changes
      # and make them at the end once we know what all has to be done.
      # However, this is easier for now, and we don't often do mass VLAN
      # changes anyway!?)
      set vbl {}
      # Pass 1: scan all the untagged lists, and if any include the
      # port and shouldn't then clear them down and queue a vb.
      foreach t [lsort -integer [array names vlanStaticUntagged]] {
         if {$t != $untagged} {
            # This one shouldn't be untagged; test whether it is
            if [qBtest $vlanStaticUntagged($t) $index] {
               #puts "$t: untagged but shouldn't be"
               set vlanStaticUntagged($t) \
                  [qBclear $vlanStaticUntagged($t) $index]
               lappend vbl \
                  [list "Q-BRIDGE-MIB!dot1qVlanStaticUntaggedPorts.$t" \
                        $vlanStaticUntagged($t)]
            }
            # Else it wasn't untagged
         }
         # Else we do want it to be untagged
      }
      # Pass 2: scan all the egress lists to make sure that everything that
      # should be there is, and nothing that shouldn't be is.
      foreach t [lsort -integer [array names vlanStaticEgress]] {
         if [qBtest $vlanStaticEgress($t) $index] {
            # The egress bit was set.  Should it have been?
            #puts "$index: egress set for tag $t"
            if {$untagged != $t && [lsearch -exact $tagged $t] < 0} {
               # No, clear it
               set vlanStaticEgress($t) [qBclear $vlanStaticEgress($t) $index]
               lappend vbl \
                  [list "Q-BRIDGE-MIB!dot1qVlanStaticEgressPorts.$t" \
                        $vlanStaticEgress($t)]
            }
            # Else it should have been, so leave it
         } else {
            # The egress bit wasn't set.  Should it have been?
            #puts "$index: egress not set for tag $t"
            if {$untagged == $t || [lsearch -exact $tagged $t] >= 0} {
               # Yes, set it
               set vlanStaticEgress($t) [qBset $vlanStaticEgress($t) $index]
               lappend vbl \
                  [list "Q-BRIDGE-MIB!dot1qVlanStaticEgressPorts.$t" \
                        $vlanStaticEgress($t)]
            }
            # Else it shouldn't have been, so leave it
         }
      }
      # Pass 3: make sure the correct untagged bit is set, and the PVID
      if {![qBtest $vlanStaticUntagged($untagged) $index]} {
         # Should have been set untagged, but wasn't
         #puts "$untagged: not untagged"
         set vlanStaticUntagged($untagged) \
            [qBset $vlanStaticUntagged($untagged) $index]
         lappend vbl \
            [list "Q-BRIDGE-MIB!dot1qVlanStaticUntaggedPorts.$untagged" \
                  $vlanStaticUntagged($untagged)]
         lappend vbl [list "Q-BRIDGE-MIB!dot1qPvid.$index" $untagged]
         # While we're here, make sure that VLAN filtering is turned on
         lappend vbl [list "Q-BRIDGE-MIB!dot1qPortIngressFiltering.$index" \
                           "true"]
      }
      #puts "vbl: $vbl"
   } else {
      # HP-VLAN
      if {![info exists vlanTagToIndex($untagged)]} {
         puts stderr "VLAN tag $untagged not in tables"
         return 0
      }
      set i $vlanTagToIndex($untagged)
      set vbl [list [list "HP-VLAN!hpVlanMemberTagged2.$i.$index" "untagged"]]
      set used $i
      foreach t $tagged {
         if {![info exists vlanTagToIndex($t)]} {
            puts stderr "VLAN tag $t not in tables"
            return 0
         }
         set i $vlanTagToIndex($t)
         lappend vbl [list "HP-VLAN!hpVlanMemberTagged2.$i.$index" "tagged"]
         lappend used $i
      }
      foreach i [lsort -integer [array names vlanIndexToTag]] {
         if {[lsearch -exact $used $i] < 0} {
            lappend vbl [list "HP-VLAN!hpVlanMemberTagged2.$i.$index" "no"]
         }
      }
   }
   if [catch {$snmpS set $vbl} why] {
      puts "snmpSetVlans: $why"
      return 0
   } else {
      puts "OK!"
      return 1
   }
}

####################
# MAU
####################

proc snmpMauAutoType {i} {
   global snmpMauAutoTypes
   global snmpS
   if {! [info exists snmpMauTypes($i)]} {
      set vbl "MAU-MIB!ifMauType.$i
               MAU-MIB!ifMauAutoNegAdminStatus.$i"
      set vbl [$snmpS getnext $vbl]
      set snmpMauAutoTypes($i) [snmp value $vbl]
   }
   set snmpMauAutoTypes($i)
}

proc snmpMauType {i} {
   lindex [snmpMauAutoType $i] 0
}

proc snmpMauAuto {i} {
   lindex [snmpMauAutoType $i] 1
}

proc snmpMauAvailableExits {i} {
   global snmpS
   set vb "MAU-MIB!ifMauMediaAvailableStateExits.$i"
   set vb [$snmpS getnext $vb]
   snmp value $vb
}

#proc snmpSetSpeed {index name speed} {
#   global snmpS
#   puts "snmpSetSpeed: index $index, name $name, speed \"$speed\""
#   if {[string compare $speed "-"] == 0} {
#      set vbl [list \
#            [list "MAU-MIB!ifMauAutoNegAdminStatus.$index.1" "enabled"]]
#   } else {
#      set vbl [list \
#            [list "MAU-MIB!ifMauAutoNegAdminStatus.$index.1" "disabled"]]
#      switch -exact -- $speed {
#         "10HD"   { set speed "MAU-MIB!dot3MauType10BaseTHD" }
#         "10FD"   { set speed "MAU-MIB!dot3MauType10BaseTFD" }
#         "100HD"  { set speed "MAU-MIB!dot3MauType100BaseTXHD" }
#         "100FD"  { set speed "MAU-MIB!dot3MauType100BaseTXFD" }
#         default  { puts "snmpSetSpeed: unknown speed $speed" ; return 0 }
#      }
#      lappend vbl [list "MAU-MIB!ifMauDefaultType.$index.1" "$speed"]
#   }
#   foreach v $vbl { puts "vbl: $v" }
#   if [catch {$snmpS set $vbl} why] {
#foreach w $why {puts "$w"}
#      #puts "snmpSetSpeed: $why"
#      return 0
#   } else {
#      puts "OK!"
#      return 1
#   }
#}

proc snmpSetSpeed {index name speed} {
   global snmpS
   puts "snmpSetSpeed: index $index, name $name, speed \"$speed\""
   switch -exact -- $speed {
      "-"      { set speed "auto-neg" }
      "10HD"   { set speed "half-duplex-10Mbits" }
      "10FD"   { set speed "full-duplex-10Mbits" }
      "100HD"  { set speed "half-duplex-100Mbits" }
      "100FD"  { set speed "full-duplex-100Mbits" }
      default  { puts "snmpSetSpeed: unknown speed $speed" ; return 0 }
   }
   set vbl [list [list "CONFIG-MIB!hpSwitchPortFastEtherMode.$index" "$speed"]]
   if [catch {$snmpS set $vbl} why] {
      puts "snmpSetSpeed: $why"
      return 0
   } else {
      puts "OK!"
      return 1
   }
}

##########
# STP
##########

proc snmpStpAdminStatus {} {
   global snmpS
   set vb "CONFIG-MIB!hpSwitchStpAdminStatus.1"
   set vb [$snmpS get $vb]
   if {[string compare [snmp value $vb] "enable"] == 0} {
      return 1
   } else {
      return 0
   }
}

proc snmpStpInfo {} {
   global snmpS
   set vbl "BRIDGE-MIB!dot1dStpPriority.0
            BRIDGE-MIB!dot1dStpRootCost.0
            BRIDGE-MIB!dot1dStpRootPort.0
            BRIDGE-MIB!dot1dStpDesignatedRoot.0"
   set vbl [$snmpS get $vbl]
   snmp value $vbl
}

# XXX: We really shouldn't assume that the mapping between
# box port IDs and dot1d IDs is null.
proc snmpStpPortState {pi} {
   global snmpS
   set vb "BRIDGE-MIB!dot1dStpPortState.$pi"
   set vb [$snmpS get $vb]
   snmp value $vb
}

proc snmpStpPortCost {pi} {
   global snmpS
   set vb "BRIDGE-MIB!dot1dStpPortPathCost.$pi"
   set vb [$snmpS get $vb]
   snmp value $vb
}

proc snmpStpPortMode {pi} {
   global snmpS
   set vb "CONFIG-MIB!hpSwitchStpPortMode.$pi"
   set vb [$snmpS get $vb]
   snmp value $vb
}

set snmpValidStpModes {"normal" "fast"}

proc snmpSetStpPort {pi mode {cost -1}} {
   global snmpS
   global snmpValidStpModes
   puts "snmpSetStpPort: $pi -> $mode/$cost"
   set mode [string tolower $mode]
   if {[lsearch -exact $snmpValidStpModes $mode] < 0} {
      puts stderr "snmpSetStpPort: bad mode $mode for port $pi"
      return 0
   }
   set vbl [list [list "CONFIG-MIB!hpSwitchStpPortMode.$pi" "$mode"]]
   if {$cost > 0} {
      lappend vbl [list "BRIDGE-MIB!dot1dStpPortPathCost.$pi" $cost]
   }
   if [catch {$snmpS set $vbl} why] {
      puts "snmpSetStpPort: $why"
      return 0
   }
   puts "OK!"
   return 1
}

##########
# Fault-finder
##########

proc snmpFfInfo {} {
   global snmpS
   set fList {}
   set ffRoot [mib oid "HP-ICF-FAULT-FINDER-MIB!hpicfFfLogCreateTime"]
   set vbl "$ffRoot.0
            HP-ICF-FAULT-FINDER-MIB!hpicfFfLogPhysEntity.0
            HP-ICF-FAULT-FINDER-MIB!hpicfFfLogFaultType.0
            HP-ICF-FAULT-FINDER-MIB!hpicfFfLogStatus.0"
   while {1} {
      set vbl [$snmpS getnext $vbl]
      if {! [mib subtree $ffRoot [snmp oid $vbl 0]]} break
      lappend fList [snmp value $vbl]
   }
   set fList
}

##########
# RMON
##########

proc snmpRmonEtherErrors {i} {
   global snmpS
   set vbl "RMON-MIB!etherStatsDropEvents.$i
            RMON-MIB!etherStatsCRCAlignErrors.$i
            RMON-MIB!etherStatsUndersizePkts.$i
            RMON-MIB!etherStatsOversizePkts.$i
            RMON-MIB!etherStatsFragments.$i
            RMON-MIB!etherStatsJabbers.$i"
   set vbl [$snmpS get $vbl]
   snmp value $vbl
}

proc snmpRmonEtherErrorLabels {} {
   return {Drop CRC Under Over Frag Jabber}
}

proc snmpRmonPortCounters {i} {
   global snmpS
   set vbl "RMON-MIB!etherStatsOctets.$i
            RMON-MIB!etherStatsPkts.$i
            RMON-MIB!etherStatsBroadcastPkts.$i
            RMON-MIB!etherStatsMulticastPkts.$i
            RMON-MIB!etherStatsCollisions.$i
            RMON-MIB!etherStatsPkts64Octets.$i
            RMON-MIB!etherStatsPkts65to127Octets.$i
            RMON-MIB!etherStatsPkts128to255Octets.$i
            RMON-MIB!etherStatsPkts256to511Octets.$i
            RMON-MIB!etherStatsPkts512to1023Octets.$i
            RMON-MIB!etherStatsPkts1024to1518Octets.$i"
   set vbl [$snmpS get $vbl]
   snmp value $vbl
}

proc snmpRmonPortLabels {} {
   return {Octets Pkts BCast MCast Coll \
           0-64 65-127 127-255 256-511 512-1023 1024-1518}
}

proc snmpRmonPortSizeCounters {i} {
   global snmpS
   set vbl "RMON-MIB!etherStatsPkts64Octets.$i
            RMON-MIB!etherStatsPkts65to127Octets.$i
            RMON-MIB!etherStatsPkts128to255Octets.$i
            RMON-MIB!etherStatsPkts256to511Octets.$i
            RMON-MIB!etherStatsPkts512to1023Octets.$i
            RMON-MIB!etherStatsPkts1024to1518Octets.$i"
   set vbl [$snmpS get $vbl]
   snmp value $vbl
}

proc snmpRmonPortSizeLabels {} {
   return {0-64 65-127 127-255 256-511 512-1023 1024-1518}
}

##########
# HP chassis
##########

proc snmpHpChassisSensors {} {
   global snmpS
   set sList {}
   set sRoot [mib oid "HP-ICF-CHASSIS!hpicfSensorDescr"]
   set vbl "$sRoot.0
            HP-ICF-CHASSIS!hpicfSensorStatus.0
            HP-ICF-CHASSIS!hpicfSensorWarnings.0
            HP-ICF-CHASSIS!hpicfSensorFailures.0"
   while {1} {
      set vbl [$snmpS getnext $vbl]
      if {! [mib subtree $sRoot [snmp oid $vbl 0]]} break
      lappend sList [snmp value $vbl]
   }
   set sList
}

##########
# HP "http" stuff
##########

proc snmpHpHttpStuff {} {
   global snmpS
   set vbl "HP-httpManageable-MIB!hpHttpMgHealth.0
            HP-httpManageable-MIB!hpHttpMgSerialNumber.0
            HP-httpManageable-MIB!hpHttpMgAssetNumber.0"
   set vbl [$snmpS get $vbl]
   snmp value $vbl
}

##########
# Port forwarding table
##########

proc snmpPortFwdTable {} {
   global snmpS
   set sList {}
   set sRoot [mib oid "STATISTICS-MIB!hpSwitchPortFdbAddress"]
   set vbl "$sRoot.0"
   while {1} {
      set vbl [$snmpS getnext $vbl]
      if {! [mib subtree $sRoot [snmp oid $vbl]]} break
      if [regexp "$sRoot\.(\[0-9\]+)\.*" [snmp oid $vbl] all index] {
         lappend sList [list $index [snmp value $vbl]]
      } else {
         puts "snmpPortFwdTable: regexp failed for <[snmp oid $vbl]>"
      }
   }
   set sList
}

##########
# Etherlike counters
##########

proc snmpPortEtherLikeErrors {i} {
   global snmpS
   set vbl "EtherLike-MIB!dot3StatsAlignmentErrors.$i
            EtherLike-MIB!dot3StatsFCSErrors.$i
            EtherLike-MIB!dot3StatsSingleCollisionFrames.$i
            EtherLike-MIB!dot3StatsMultipleCollisionFrames.$i
            EtherLike-MIB!dot3StatsDeferredTransmissions.$i
            EtherLike-MIB!dot3StatsLateCollisions.$i
            EtherLike-MIB!dot3StatsExcessiveCollisions.$i
            EtherLike-MIB!dot3StatsInternalMacTransmitErrors.$i
            EtherLike-MIB!dot3StatsCarrierSenseErrors.$i
            EtherLike-MIB!dot3StatsFrameTooLongs.$i"
   set vbl [$snmpS get $vbl]
   snmp value $vbl
}

proc snmpPortEtherLikeErrorsLabels {} {
   return {Align FCS SColl MColl Defer Late Excsv IMacTX CSense TooLong}
}

proc snmpPortEtherLikeErrorsDispflags {} {
   return {1 1 -1 -1 -1 1 1 1 0 1}
}
